package info.textgrid.middleware;

import java.io.IOException;
import java.io.StringWriter;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBElement;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.datatype.DatatypeConfigurationException;
import javax.xml.datatype.XMLGregorianCalendar;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import info.textgrid.middleware.oaipmh.GetRecordType;
import info.textgrid.middleware.oaipmh.IdentifyType;
import info.textgrid.middleware.oaipmh.ListIdentifiersType;
import info.textgrid.middleware.oaipmh.ListMetadataFormatsType;
import info.textgrid.middleware.oaipmh.ListRecordsType;
import info.textgrid.middleware.oaipmh.ListSetsType;
import info.textgrid.middleware.oaipmh.OAIPMHtype;
import info.textgrid.middleware.oaipmh.ObjectFactory;
import info.textgrid.middleware.oaipmh.RequestType;
import info.textgrid.middleware.oaipmh.VerbType;

/**
 * <p>
 * This class takes all possible arguments of an OAI-PMH request and proceeds depended on the given
 * arguments. It also deals as response handler.
 * </p>
 * 
 * @author Maximilian Brodhun, SUB Göttingen
 * @author Stefan E. Funk, SUB Göttingen
 * @version 2021-07-06
 * @since 2014-01-29
 */
public class OAIPMHImpl implements OAIPMHProducer {

  // NOTE Changed OpenAireIdentifierList into IdentifierListDelivererDATACITE!

  // **
  // STATICS
  // **

  private static ObjectFactory obf = new ObjectFactory();

  // **
  // PRIVATES
  // **

  private ErrorHandler error = new ErrorHandler();
  private Log log = LogFactory.getLog(OAIPMHImpl.class);
  private RepIdentification rep;

  private RecordListDelivererDC recordListDC;
  private RecordListDelivererIDIOM recordListIDIOM;
  // private OpenAireRecord openAireRecord;
  private OpenAireRecordList openAireRecordList;
  // private OpenAireIdentifierList openAireIdentifierList;

  private RecordDelivererDC recDelivDC;
  private RecordDelivererIDIOM recDelivIDIOM;
  private RecordDelivererDATACITE recDelivDATACITE;

  private IdentifierListDelivererDC identifierListDC;
  private IdentifierListDelivererIDIOM identifierListIDIOM;
  private IdentifierListDelivererDATACITE identifierListDATACITE;

  private MetadataFormatListDelivererInterface metadataFormatList;
  private SetDeliverer setList;

  // **
  // PUBLIC
  // **

  public boolean textgrid;
  public boolean dariah;

  /**
   * @param rep
   * @param recDelivDC
   * @param recDelivIDIOM
   * @param recordListDC
   * @param recordListIDIOM
   * @param metadataFormatList
   * @param setList
   * @param identifierList
   * @param identifierListIDIOM
   */
  public OAIPMHImpl(
      RepIdentification rep, // 0
      RecordDelivererDC recDelivDC, // 1
      RecordDelivererIDIOM recDelivIDIOM, // 2
      // OpenAireRecord openAireRecord, // 3
      RecordDelivererDATACITE recDelivDATACITE, // 3
      RecordListDelivererDC recordListDC, // 4
      RecordListDelivererIDIOM recordListIDIOM, // 5
      OpenAireRecordList openAireRecordList, // 6
      MetadataFormatListDelivererInterface metadataFormatList, // 7
      SetDeliverer setList, // 8
      IdentifierListDelivererDC identifierList, // 9
      IdentifierListDelivererIDIOM identifierListIDIOM, // 10
      // OpenAireIdentifierList openAireIdentifierList // 11
      IdentifierListDelivererDATACITE identifierListDATACITE // 11
  ) {
    this.rep = rep;
    this.recDelivDC = recDelivDC;
    this.recDelivIDIOM = recDelivIDIOM;
    // this.openAireRecord = openAireRecord;
    this.recDelivDATACITE = recDelivDATACITE;
    this.recordListDC = recordListDC;
    this.recordListIDIOM = recordListIDIOM;
    this.openAireRecordList = openAireRecordList;
    this.metadataFormatList = metadataFormatList;

    this.log
        .debug("MetadataFormatList Implementing Class: " + metadataFormatList.getClass().getName());

    this.setList = setList;
    this.identifierListDC = identifierList;
    this.identifierListIDIOM = identifierListIDIOM;
    // this.openAireIdentifierList = openAireIdentifierList;
    this.identifierListDATACITE = identifierListDATACITE;
  }

  /*
   * (non-Javadoc)
   * 
   * @see info.textgrid.middleware.OAIPMHProducer#postRequest(java.lang.String, java.lang.String,
   * java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String)
   */
  public String postRequest(String verb, String identifier, String metadataPrefix, String set,
      String from, String until, String resumptionToken) {

    String result = "";

    try {
      result = getStringFromJAXBOAIElement(verb,
          handleRequest(verb, identifier, metadataPrefix, set, from, until, resumptionToken));

    } catch (ParseException | IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }

    return result;
  }

  /*
   * (non-Javadoc)
   * 
   * @see info.textgrid.middleware.OAIPMHProducer#getRequest(java.lang.String, java.lang.String,
   * java.lang.String, java.lang.String, java.lang.String, java.lang.String, java.lang.String)
   */
  public String getRequest(String verb, String identifier, String metadataPrefix,
      String set, String from, String until, String resumptionToken) {

    String result = "";

    try {
      result = getStringFromJAXBOAIElement(verb,
          handleRequest(
              verb,
              identifier,
              metadataPrefix,
              set,
              from,
              until,
              resumptionToken));

    } catch (ParseException | IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }

    return result;
  }

  /**
   * @return
   */
  public String getVersion() {
    return OaipmhServiceVersion.BUILDNAME + "-" + OaipmhServiceVersion.VERSION + "+"
        + OaipmhServiceVersion.BUILDDATE;
  }

  /**
   * <p>
   * Checks the Identify request of correctness and response including errors in case of an
   * incorrect request.
   * </p>
   * 
   * @param oaipmhRoot
   * @param request
   * @return
   */
  public OAIPMHtype identifyRequest(OAIPMHtype oaipmhRoot, RequestType request) {

    List<String> errorValues = new ArrayList<String>();
    IdentifyType tgRepIdentificationRequest = new IdentifyType();

    if (this.rep.requestChecker(request)) {

      tgRepIdentificationRequest.setBaseURL(this.rep.getBaseUrl());
      tgRepIdentificationRequest.setDeletedRecord(this.rep.getDeletedRecordStatus());
      tgRepIdentificationRequest.setEarliestDatestamp(this.rep.getEarliestDatestamp());
      tgRepIdentificationRequest.setGranularity(this.rep.getGranularity());
      tgRepIdentificationRequest.setRepositoryName(this.rep.getRepositoryName());
      tgRepIdentificationRequest.setProtocolVersion(this.rep.getProtocolVersion());
      tgRepIdentificationRequest.getAdminEmail().add(this.rep.getAdminMaiAddresss());
      oaipmhRoot.setIdentify(tgRepIdentificationRequest);

    } else {
      if (request.getIdentifier() != null) {
        errorValues.add("identifier");
      }
      if (request.getFrom() != null) {
        errorValues.add("from");
      }
      if (request.getMetadataPrefix() != null) {
        errorValues.add("metadataPrefix");
      }
      if (request.getResumptionToken() != null) {
        errorValues.add("ResumptionToken");
      }
      if (request.getUntil() != null) {
        errorValues.add("Until");
      }
      if (request.getSet() != null) {
        errorValues.add("Set");
      }

      this.error.setError(TGConstants.OAI_BAD_ARGUMENT,
          "The request includes illegal arguments: " + errorValues);
      oaipmhRoot.getError().add(this.error.getError());
    }

    return oaipmhRoot;
  }

  /**
   * <p>
   * Checks the GetRecord request of correctness and response including errors in case of an
   * incorrect request.
   * </p>
   * 
   * @param oaipmhRoot
   * @param request
   * @return
   */
  public OAIPMHtype getRecordRequest(OAIPMHtype oaipmhRoot, RequestType request) {

    // Check request. In GetRecord: allowed parameters: identifier, metadataPrefix, verb forbidden:
    // resumptionToken, from, till, set optional: no one.

    ErrorHandler requestErrors = RecordDelivererAbstract.requestChecker(request);

    if (requestErrors.getError().getCode() != null) {
      oaipmhRoot.getError().add(requestErrors.getError());
    }

    // Get Record Request is done HERE!
    else {

      // Default is DC.
      RecordDelivererInterface recDeliv = this.recDelivDC;

      // Take IDIOM if IDIOM prefix.
      if (request.getMetadataPrefix().equals(TGConstants.METADATA_IDIOM_PREFIX)) {
        recDeliv = this.recDelivIDIOM;
      }
      if (request.getMetadataPrefix().equals(TGConstants.METADATA_OPENAIRE_PREFIX)) {
        recDeliv = this.recDelivDATACITE;
      }

      // Finally start the QUERY!
      GetRecordType getRecord = new GetRecordType();

      String id = request.getIdentifier();
      try {
        getRecord = recDeliv.getRecordById(id);
      } catch (ParseException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      } catch (DatatypeConfigurationException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }

      if (getRecord != null) {
        oaipmhRoot.setGetRecord(getRecord);
      } else {
        ErrorHandler idError = new ErrorHandler();
        idError.setError(TGConstants.OAI_ID_DOES_NOT_EXIST, "No matching identifier ");
        oaipmhRoot.getError().add(idError.getError());
      }
    }

    return oaipmhRoot;
  }

  /**
   * <p>
   * Checks the ListIdentifiers request of correctness and response including errors in case of an
   * incorrect request.
   * </p>
   * 
   * @param oai
   * @param request
   * @return
   * @throws ParseException
   */
  public OAIPMHtype listIdentifiersRequest(OAIPMHtype oaipmhRoot, RequestType request)
      throws ParseException {

    ErrorHandler requestErrors = IdentifierListDelivererAbstract.requestChecker(request);

    if (requestErrors.getError().getValue() != null) {
      oaipmhRoot.getError().add(requestErrors.getError());
    } else {

      // If metadataFormat IS SET, set recordListDeliverer accordingly.
      IdentifierListDelivererInterface idListDeliv = null;
      if (request.getMetadataPrefix() != null) {
        if (request.getMetadataPrefix().equals(OAIPMHUtilities.OAIDC_PREFIX)) {
          idListDeliv = this.identifierListDC;
        } else if (request.getMetadataPrefix().equals(TGConstants.METADATA_IDIOM_PREFIX)) {
          idListDeliv = this.identifierListIDIOM;
        } else if (request.getMetadataPrefix().equals(TGConstants.METADATA_OPENAIRE_PREFIX)) {
          // idListDeliv = this.openAireIdentifierList;
          idListDeliv = this.identifierListDATACITE;
        }
      }

      // If metadata format IS NOT SET, check resumption token hash maps to decide if we have to
      // serve oai_dc or idiom_mets metadata format.
      else {
        boolean restokDCExisting = IdentifierListDelivererDC.cursorCollector != null
            && IdentifierListDelivererDC.cursorCollector.containsKey(request.getResumptionToken());
        boolean restokIDIOMExisting = IdentifierListDelivererIDIOM.cursorCollector != null
            && IdentifierListDelivererIDIOM.cursorCollector
                .containsKey(request.getResumptionToken());
        boolean restokOpenAireExisting = OAIPMHUtilities.cursorCollector != null
            && IdentifierListDelivererAbstract.cursorCollector
                .containsKey(request.getResumptionToken());
        if (restokDCExisting) {
          idListDeliv = this.identifierListDC;
        } else if (restokIDIOMExisting) {
          idListDeliv = this.identifierListIDIOM;
        } else if (restokOpenAireExisting) {
          // idListDeliv = this.openAireIdentifierList;
          idListDeliv = this.identifierListDATACITE;
        }
      }

      ListIdentifiersType listIdentifiers = null;
      try {
        this.log.debug("GET FROM: " + request.getFrom());
        this.log.debug("GET UNTIL: " + request.getUntil());

        listIdentifiers = idListDeliv.processIdentifierList(request.getFrom(), request.getUntil(),
            request.getSet(), request.getResumptionToken());
      } catch (IOException e) {
        // TODO Auto-generated catch block
        e.printStackTrace();
      }

      if (listIdentifiers != null) {
        oaipmhRoot.setListIdentifiers(listIdentifiers);
      } else {
        ErrorHandler idError = new ErrorHandler();
        idError.setError(TGConstants.OAI_NO_RECORD_MATCH, "The value of the identifier: "
            + request.getIdentifier() + " is unknown or illegal in this repository");
        oaipmhRoot.getError().add(idError.getError());
      }
    }

    return oaipmhRoot;
  }

  /**
   * <p>
   * Checks the ListSets request of correctness and response including errors in case of an
   * incorrect request.
   * </p>
   * 
   * @param oai
   * @param request
   * @return
   */
  public OAIPMHtype listSetsRequest(OAIPMHtype oaipmhRoot, RequestType request) {

    ListSetsType sl = this.setList.setListBuilder();

    if (this.setList.requestChecker(request)) {
      oaipmhRoot.setListSets(sl);
    } else {
      List<String> errorValues = new ArrayList<String>();
      ErrorHandler badArgument = new ErrorHandler();
      if (request.getIdentifier() != null) {
        errorValues.add("identifier");
      }
      if (request.getFrom() != null) {
        errorValues.add("from");
      }
      if (request.getMetadataPrefix() != null) {
        errorValues.add("metadataPrefix");
      }
      if (request.getResumptionToken() != null) {
        errorValues.add("ResumptionToken");
      }
      if (request.getUntil() != null) {
        errorValues.add("until");
      }
      if (request.getSet() != null) {
        errorValues.add("set");
      }

      badArgument.setError(TGConstants.OAI_BAD_ARGUMENT,
          "The request includes illegal arguments: " + errorValues);
      oaipmhRoot.getError().add(badArgument.getError());
    }

    return oaipmhRoot;
  }

  /**
   * @param id
   * @param oaipmhRoot
   * @param request
   * @return
   */
  public OAIPMHtype listMetadataFormatsRequest(String id, OAIPMHtype oaipmhRoot,
      RequestType request) {

    this.log.debug("request: " + request.getValue());
    this.log.debug("id: " + id);

    List<String> errorValues = new ArrayList<String>();
    ListMetadataFormatsType listMF = new ListMetadataFormatsType();

    if (this.metadataFormatList.requestChecker(request)) {

      if (id.isEmpty()) {
        listMF = this.metadataFormatList.setMetadataFormatList();
      } else {
        listMF = this.metadataFormatList.setMetadataFormatList(request.getIdentifier());
      }

      if (listMF != null) {
        oaipmhRoot.setListMetadataFormats(listMF);
      } else {
        this.error.setError(TGConstants.OAI_NO_RECORD_MATCH, "The value of the identifier: "
            + request.getIdentifier() + " is unknown or illegal in this repository");
        oaipmhRoot.getError().add(this.error.getError());
      }
    }

    else {

      if (request.getFrom() != null) {
        errorValues.add("from");
      }
      if (request.getMetadataPrefix() != null) {
        errorValues.add("metadataPrefix");
      }
      if (request.getResumptionToken() != null) {
        errorValues.add("resumptionToken");
      }
      if (request.getSet() != null) {
        errorValues.add("set");
      }
      if (request.getUntil() != null) {
        errorValues.add("until");
      }

      this.error.setError(TGConstants.OAI_BAD_ARGUMENT, "The request includes illegal "
          + "arguments or is missing required arguments: " + errorValues);
      oaipmhRoot.getError().add(this.error.getError());
    }

    return oaipmhRoot;
  }

  /**
   * @param oai
   * @param request
   * @return
   * @throws ParseException
   * @throws IOException
   */
  public OAIPMHtype listRecordsRequest(OAIPMHtype oaipmhRoot, RequestType request)
      throws ParseException, IOException {

    // Check request's syntax.
    ErrorHandler requestErrors = RecordListDelivererAbstract.requestChecker(request);

    if (requestErrors.getError().getCode() != null) {
      oaipmhRoot.getError().add(requestErrors.getError());
    } else {

      // If metadataFormat IS SET, set recordListDeliverer accordingly.
      RecordListDelivererInterface recListDeliv = null;
      if (request.getMetadataPrefix() != null) {
        if (request.getMetadataPrefix().equals(OAIPMHUtilities.OAIDC_PREFIX)) {
          recListDeliv = this.recordListDC;
        }
        if (request.getMetadataPrefix().equals(TGConstants.METADATA_IDIOM_PREFIX)) {
          recListDeliv = this.recordListIDIOM;
        }
        if (request.getMetadataPrefix().equals(TGConstants.METADATA_OPENAIRE_PREFIX)) {
          recListDeliv = this.openAireRecordList;
        }
      }

      // If metadata format IS NOT SET, check resumption token hash maps to decide if we have to
      // serve oai_dc or idiom_mets metadata format.
      else {
        boolean restokDCExisting = RecordListDelivererDC.cursorCollector != null
            && RecordListDelivererDC.cursorCollector.containsKey(request.getResumptionToken());
        boolean restokIDIOMExisting = RecordListDelivererIDIOM.cursorCollector != null
            && RecordListDelivererIDIOM.cursorCollector.containsKey(request.getResumptionToken());

        boolean restokOpenAireExisting = RecordListDelivererAbstract.cursorCollector != null
            && RecordListDelivererAbstract.cursorCollector
                .containsKey(request.getResumptionToken());

        if (restokDCExisting) {
          recListDeliv = this.recordListDC;
        } else if (restokIDIOMExisting) {
          recListDeliv = this.recordListIDIOM;
        } else if (restokOpenAireExisting) {
          recListDeliv = this.openAireRecordList;
        }
      }

      ListRecordsType listRecords = recListDeliv.getRecords(
          request.getFrom(),
          request.getUntil(),
          request.getSet(),
          request.getResumptionToken());

      if (listRecords != null) {
        /*
         * if (this.recordListDC.getResultSize() == 0) { requestErrors.setError("RecordMatchError",
         * "The combination of the values of the from, until, set and metadataPrefix arguments results in an empty list."
         * ); oaipmhRoot.getError().add(requestErrors.getError()); } else {
         */
        oaipmhRoot.setListRecords(listRecords);
        // }
      }
    }

    return oaipmhRoot;
  }

  /**
   * <p>
   * Taking the string from the REST request and converting into an OAIPMH verb type.
   * </p>
   * 
   * @param verb : what is to within the repository
   * @return verbParam: the verbParam as an OAIPMH verb type
   */
  public VerbType setVerb(String verb) {

    VerbType verbParam = null;

    if (verb != null && verb.equals("Identify")) {
      verbParam = VerbType.IDENTIFY;
    } else if (verb != null && verb.equals("ListMetadataFormats")) {
      verbParam = VerbType.LIST_METADATA_FORMATS;
    } else if (verb != null && verb.equals("ListSets")) {
      verbParam = VerbType.LIST_SETS;
    } else if (verb != null && verb.equals("ListIdentifiers")) {
      verbParam = VerbType.LIST_IDENTIFIERS;
    } else if (verb != null && verb.equals("ListRecords")) {
      verbParam = VerbType.LIST_RECORDS;
    } else if (verb != null && verb.equals("GetRecord")) {
      verbParam = VerbType.GET_RECORD;
    } else if (verb != null && verb.length() > 0) {
      this.error.setError(TGConstants.OAI_VERB_ERROR, "Illegal OAI verb");
    }
    // Verb not set --> please do use verb=Identify!
    // FIXME Is the following appropriate?? Do we miss any error messages that way??
    else {
      verbParam = VerbType.IDENTIFY;
    }

    return verbParam;
  }

  /**
   * <p>
   * Setting the from value for the request.
   * </p>
   * 
   * @param from
   */
  public void setFromRequestValue(String from, RequestType request) {
    if (!from.isEmpty()) {
      request.setFrom(from);
    }
  }

  /**
   * <p>
   * Setting the until value for the request.
   * </p>
   * 
   * @param until
   */
  public void setUntilRequestValue(String until, RequestType request) {
    if (!until.isEmpty()) {
      request.setUntil(until);
    }
  }

  /**
   * <p>
   * Setting the identifier value for the request.
   * </p>
   * 
   * @param identifier
   */
  public void setIdentifierRequestValue(String identifier, RequestType request) {
    if (!identifier.isEmpty()) {
      request.setIdentifier(identifier);
    }
  }

  /**
   * <p>
   * Setting the metadataPrefix value for the request.
   * </p>
   * 
   * @param metadataPrefix
   */
  public void setMetadataPrefixRequestValue(String metadataPrefix, RequestType request) {
    if (!metadataPrefix.isEmpty()) {
      request.setMetadataPrefix(metadataPrefix);
    }
  }

  /**
   * <p>
   * Setting the resumptionToken value for the request.
   * </p>
   * 
   * @param resumptionToken
   */
  public void setResumptionTokenRequestValue(String resumptionToken, RequestType request) {
    if (!resumptionToken.isEmpty()) {
      request.setResumptionToken(resumptionToken);
    }
  }

  /**
   * <p>
   * Setting the set value for the request
   * </p>
   * 
   * @param set
   * @param request
   */
  public void setSetRequestValue(String set, RequestType request) {
    if (!set.isEmpty()) {
      request.setSet(set);
    }
  }

  /**
   * @param textgrid
   */
  public void setTextGrid(boolean textgrid) {
    this.textgrid = textgrid;
  }

  /**
   * @param dariah
   */
  public void setDariah(boolean dariah) {
    this.dariah = dariah;
  }

  // **
  // PRIVATE METHODS
  // **

  /**
   * @param verb
   * @param identifier
   * @param metadataPrefix
   * @param set
   * @param from
   * @param until
   * @param resumptionToken
   * @return
   * @throws ParseException
   * @throws IOException
   */
  private JAXBElement<OAIPMHtype> handleRequest(String verb, String identifier,
      String metadataPrefix, String set, String from, String until, String resumptionToken)
      throws ParseException, IOException {

    JAXBElement<OAIPMHtype> result;

    OAIPMHtype response = new OAIPMHtype();
    VerbType verbParam = setVerb(verb);
    RequestType request = new RequestType();

    // Set the responseDate of today.
    try {
      XMLGregorianCalendar convertedDateFormat;
      convertedDateFormat = OAIPMHUtilities
          .convertDateFormat(OAIPMHUtilities.getXMLGregorianCalendarNow().toString());
      response.setResponseDate(convertedDateFormat);
    } catch (ParseException e) {
      this.log.debug("can not parse date format");
    } catch (DatatypeConfigurationException e) {
      this.log.debug("datatype configuration failed");
    }

    // Convert the request values.
    setFromRequestValue(from, request);
    setUntilRequestValue(until, request);
    setIdentifierRequestValue(identifier, request);
    setMetadataPrefixRequestValue(metadataPrefix, request);
    setResumptionTokenRequestValue(resumptionToken, request);
    setSetRequestValue(set, request);
    request.setVerb(verbParam);

    if (this.textgrid == true) {
      request.setValue(TGConstants.TG_REP_BASEURL);
    }
    if (this.dariah == true) {
      request.setValue(DARIAHConstants.DARIAH_REP_BASEURL);
    }

    if (verbParam != null) {
      if (verbParam.value().equals("Identify")) {
        response = identifyRequest(response, request);
      } else if (verbParam.value().equals("GetRecord")) {
        response = getRecordRequest(response, request);
      } else if (verbParam.value().equals("ListIdentifiers")) {
        response = listIdentifiersRequest(response, request);
      } else if (verbParam.value().equals("ListSets")) {
        response = listSetsRequest(response, request);
      } else if (verbParam.value().equals("ListMetadataFormats")) {
        response = listMetadataFormatsRequest(identifier, response, request);
      } else if (verbParam.value().equals("ListRecords")) {
        response = listRecordsRequest(response, request);
      }
    } else {
      if (this.error.getError().getValue() != null
          && this.error.getError().getValue().equals("Illegal OAI verb")) {
        this.error.setError(TGConstants.OAI_VERB_ERROR, "Illegal OAI verb");
        response.getError().add(this.error.getError());
      } else {
        this.error.setError(TGConstants.OAI_VERB_ERROR, "Illegal OAI verb");
        response.getError().add(this.error.getError());
      }
    }

    response.setRequest(request);
    result = obf.createOAIPMH(response);

    return result;
  }

  /**
   * <p>
   * Transforms the JaxB element into a string that contains the xsi:schemaLocation attribute. No
   * other way has been working here! I tried EVERYTHING else! Really!
   * </p>
   * 
   * @param theVerb
   * @param theJAXBOAIElement
   * @return
   */
  public static String getStringFromJAXBOAIElement(final String theVerb,
      final JAXBElement<OAIPMHtype> theJAXBOAIElement) {

    String result = "";

    // Always add OAIPMH schema location.
    String schemaLocations =
        OAIPMHUtilities.OAIPMH_NAMESPACE + " " + OAIPMHUtilities.OAIPMH_SCHEMA_LOCATION;
    // Add OAI_DC and DC schema location if needed.
    if (theVerb != null && (theVerb.equals("GetRecord") || theVerb.endsWith("ListRecords"))) {
      schemaLocations +=
          OAIPMHUtilities.OAIDC_NAMESPACE + " " + OAIPMHUtilities.OAIDC_SCHEMA_LOCATION +
              " " + OAIPMHUtilities.DC_NAMESPACE + " " + OAIPMHUtilities.DC_SCHEMA_LOCATION;
    }

    // PLEASE NOTE We always get oai_dc and dc prefixes defined in the root element due to the
    // bindings.xml file. Seems to be no problem for the time being.

    try {
      // OBACHT! GEMOKEL! Add schemaLocations to OAI root element here!
      JAXBContext jc = JAXBContext.newInstance(OAIPMHtype.class);

      Marshaller marshaller = jc.createMarshaller();
      // Comment out if one-line-output wanted!
      marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
      marshaller.setProperty(Marshaller.JAXB_SCHEMA_LOCATION, schemaLocations);

      StringWriter sw = new StringWriter();
      marshaller.marshal(theJAXBOAIElement, sw);
      result = sw.toString();
      sw.close();

    } catch (JAXBException | IOException e) {
      // TODO Auto-generated catch block
      e.printStackTrace();
    }

    return result;
  }

}
